import { Component, Input, Output, EventEmitter, OnInit, AfterViewInit } from "@angular/core";

import { MatDialog } from "@angular/material";

import {
    Pab,
    Session,
    Props,
    CalculatorType,
    Cloisons,
    Nub,
    Structure,
    ParallelStructure,
    ParamDefinition,
    round,
    CloisonAval
 } from "jalhyd";

 import { sprintf } from "sprintf-js";

import { I18nService } from "../../services/internationalisation/internationalisation.service";
import { FormulaireService } from "../../services/formulaire/formulaire.service";
import { ApplicationSetupService } from "../../services/app-setup/app-setup.service";
import { NotificationsService } from "../../services/notifications/notifications.service";
import { PabTable } from "../../formulaire/pab-table";
import { DialogEditPabComponent } from "../dialog-edit-pab/dialog-edit-pab.component";

/**
 * The big editable data grid for calculator type "Pab" (component)
 */
@Component({
    selector: "pab-table",
    templateUrl: "./pab-table.component.html",
    styleUrls: [
        "./pab-table.component.scss"
    ]
})
export class PabTableComponent implements AfterViewInit, OnInit {

    @Input()
    private pabTable: PabTable;

    /** flag de validité des FieldSet enfants */
    private _isValid = false;

    /** événément de changement de validité */
    @Output()
    private validChange = new EventEmitter();

    /** événément de changement de valeur d'un input */
    @Output()
    private inputChange = new EventEmitter();

    /** underlying Pab, binded to the rows */
    private model: Pab;

    /** general headers above the columns */
    public headers: any[];

    /** columns headers description */
    public cols: any[];

    /** data binded to the table */
    public rows: any[];

    /** number of children to add when clicking "add" or "clone" button */
    public childrenToAdd = 1;

    /** items currently selected */
    private selectedItems: Nub[];

    /** used for shift+click implementation */
    private latestClickedCell: any;

    public constructor(
        private i18nService: I18nService,
        private formService: FormulaireService,
        private editPabDialog: MatDialog,
        private appSetupService: ApplicationSetupService,
        private notifService: NotificationsService
    ) {
        this.selectedItems = [];
    }

    public get title(): string {
        return this.i18nService.localizeText("INFO_PAB_TABLE");
    }

    /** Global Pab validity */
    public get isValid() {
        return this._isValid;
    }

    /** returns true if the cell has an underlying model (ie. is editable) */
    public hasModel(cell: any): boolean {
        return (cell !== undefined && cell.model !== undefined);
    }

    /** returns true if the cell is an editable number */
    public isNumberInput(cell: any): boolean {
        return this.hasModel(cell) && ! this.isSelect(cell);
    }

    /** returns true if the cell is a select box @TODO rename */
    public isSelect(cell: any): boolean {
        return this.hasModel(cell) && (cell.options !== undefined);
    }

    /** value to display in a cell, if it is not editable */
    public cellValue(cell: any) {
        if (cell === undefined) {
            return "";
        } else {
            if (this.hasModel(cell)) {
                return cell.model;
            } else {
                return cell.value;
            }
        }
    }

    /** "title" tooltip to display in a cell */
    public cellTitle(cell: any) {
        if (cell !== undefined && cell.title !== undefined) {
            return cell.title;
        } else {
            return "";
        }
    }

    public rowSpan(cell: any) {
        if (cell !== undefined && cell.rowspan) {
            return cell.rowspan;
        }
        return undefined;
    }

    public colSpan(cell: any) {
        if (cell !== undefined && cell.colspan) {
            return cell.colspan;
        }
        return undefined;
    }

    /**
     * Checks that input value is a valid number, according to input[type="number"] algorithm,
     * and stores it in cell.uiValidity, so that the <td> element can access it and get angry
     * if input is invalid
     */
    public inputValueChanged($event, cell) {
        // console.log("input value changed", $event.target.validity.valid, $event.target.validity);
        if ($event && $event.target && $event.target.validity) {
            cell.uiValidity = $event.target.validity.valid;
        }
        this.updateValidity();
        // send input change event (used to reset form results)
        this.inputChange.emit();
    }

    /**
     * Returns true if current cell is bound to a model that says its input value is
     * no valid, or if characters typed in the input field are not a valid number
     * (read from cell.uiValidity, see inputValueChanged() above)
     */
    public isInvalid(cell: any): boolean {
        let valid = true;
        if (this.hasModel(cell) && cell.model instanceof ParamDefinition) {
            valid = valid && cell.model.isValid;
        }
        if (cell.uiValidity !== undefined) {
            valid = valid && cell.uiValidity;
        }
        return ! valid;
    }

    /** returns true if the cell / row has a selectable item */
    public isSelectable(cellOrRow: any): boolean {
        return (
            cellOrRow !== undefined
            && cellOrRow.selectable
        );
    }

    /** returns true if the cell / row has a selectableColumn item */
    public isSelectableByColumn(cellOrRow: any): boolean {
        return (
            cellOrRow !== undefined
            && cellOrRow.selectableColumn
        );
    }

    /**
     *  - checks if the cell / row has a selectable item, that is currently
     *      selected
     *  - if cell / row has a selectableColumn attribute, also checks if
     *      this column is selected
     *
     *  returns true if at least one criterion is met
     */
    public isSelected(cellOrRow: any): boolean {
        let cellSelected = false;
        let columnSelected = false;
        // cell
        if (this.isSelectable(cellOrRow)) {
            cellSelected = true;
            if (Array.isArray(cellOrRow.selectable)) {
                for (const elt of cellOrRow.selectable) {
                    cellSelected = cellSelected && this.selectedItems.includes(elt);
                }
            } else {
                cellSelected = this.selectedItems.includes(cellOrRow.selectable);
            }
        }
        // column
        if (this.isSelectableByColumn(cellOrRow)) {
            columnSelected = this.isDeviceColumnSelected(cellOrRow.selectableColumn);
        }
        // done
        return (cellSelected || columnSelected);
    }

    /**
     * returns true if every wall (including downwall) has its nth device
     * selected (or has no nth device)
     */
    public isDeviceColumnSelected(n: number): boolean {
        let ok = true;
        for (const c of this.model.children) {
            const nthChild = c.getChildren()[n];
            if (nthChild) {
                ok = ok && this.selectedItems.includes(nthChild);
            }
        }
        const nthChildDW = this.model.downWall.getChildren()[n];
        if (nthChildDW) {
            ok = ok && this.selectedItems.includes(nthChildDW);
        }
        return ok;
    }

    /**
     * selects or unselects the clicked cell, depending on its current state
     * and the modifier key held if any
     */
    public toggleSelection(cell: any, $event: any) {
        if (
            this.isSelectable(cell)
            && ! this.hasModel(cell) // editable cells listen to the click event for edition only
        ) {
            if ($event.shiftKey && cell !== this.latestClickedCell) { // shift + click
                // interpolate from this.latestClickedCell to this one
                if (! Array.isArray(cell.selectable)) { // multiselectable cells are not managed
                    const wallsUIDs = this.getSortedWallsUIDs();
                    let posOld: number;
                    let posNew: number;
                    // find positions depending on types, and only if types are equal
                    if (cell.selectable instanceof ParallelStructure) {
                        if (this.latestClickedCell.selectable instanceof ParallelStructure) {
                            // interpolate walls
                            posOld = wallsUIDs.indexOf(this.latestClickedCell.selectable.uid);
                            posNew = wallsUIDs.indexOf(cell.selectable.uid);
                            if (posOld !== posNew) {
                                if (posOld > posNew) {
                                    // invert order
                                    [ posOld, posNew ] = [ posNew, posOld ];
                                }
                                // go
                                for (let i = posOld; i <= posNew; i++) {
                                    if (i < this.model.children.length) {
                                        // push regular wall
                                        this.selectedItems.push(this.model.children[i]);
                                    } else {
                                        // push downwall
                                        this.selectedItems.push(this.model.downWall);
                                    }
                                    this.latestClickedCell = cell;
                                }
                            }
                        }
                    } else if (cell.selectable instanceof Structure) {
                        if (this.latestClickedCell.selectable instanceof Structure) {
                            // accept interpolation only if both devices are on the same column
                            const columnOld = this.latestClickedCell.selectable.findPositionInParent();
                            const columnNew = cell.selectable.findPositionInParent();
                            if (columnOld === columnNew) {
                                // interpolate devices
                                posOld = wallsUIDs.indexOf(this.latestClickedCell.selectable.parent.uid);
                                posNew = wallsUIDs.indexOf(cell.selectable.parent.uid);
                                if (posOld !== posNew) {
                                    if (posOld > posNew) {
                                        // invert order
                                        [ posOld, posNew ] = [ posNew, posOld ];
                                    }
                                    // go
                                    for (let i = posOld; i <= posNew; i++) {
                                        if (i < this.model.children.length) {
                                            // push regular wall
                                            this.selectedItems.push(this.model.children[i].structures[columnOld]);
                                        } else {
                                            // push downwall
                                            this.selectedItems.push(this.model.downWall.structures[columnOld]);
                                        }
                                    }
                                    this.latestClickedCell = cell;
                                }
                            }
                        }
                    }
                    // clean selected items list (deduplicate, sort)
                    this.selectedItems = this.selectedItems.filter(
                        (item, index) => this.selectedItems.indexOf(item) === index // deduplicate
                    );
                    this.sortSelectedItems();
                }

            } else if (
                $event.ctrlKey // ctrl + click
                || ($event.shiftKey && cell === this.latestClickedCell) // shift on same cell => equiv. of ctrl
            ) {
                if (this.isSelected(cell)) {
                    // unselect this cell / these cells
                    if (Array.isArray(cell.selectable)) {
                        this.selectedItems = this.selectedItems.filter(e => ! cell.selectable.includes(e));
                    } else {
                        this.selectedItems = this.selectedItems.filter(e => e !== cell.selectable);
                    }
                } else {
                    // add this cell / these cells to selection
                    if (Array.isArray(cell.selectable)) {
                        this.selectedItems = this.selectedItems.concat(cell.selectable);
                        this.selectedItems = this.selectedItems.filter(
                            (item, index) => this.selectedItems.indexOf(item) === index // deduplicate
                        );
                    } else {
                        this.selectedItems.push(cell.selectable);
                    }
                    this.sortSelectedItems();
                }
                this.latestClickedCell = cell;

            } else { // just a click
                if (this.isSelected(cell)) {
                    // select nothing
                    this.selectedItems = [];
                } else {
                    // select this cell / thses cells only
                    if (Array.isArray(cell.selectable)) {
                        this.selectedItems = cell.selectable.slice(); // array copy
                    } else {
                        this.selectedItems = [ cell.selectable ];
                    }
                }
                this.latestClickedCell = cell;
            }

            // clean list
            this.selectedItems = this.selectedItems.filter(e => e !== undefined);

            this.clearSelection();
            $event.preventDefault();
            $event.stopPropagation();
            return false;
        }
    }

    // quick getter for 1st selected item
    public get selectedItem() {
        if (this.selectedItems.length === 0) {
            throw new Error("get selectedItem() : no item selected");
        }
        return this.selectedItems[0];
    }

    // prevents Firefox to display weird cell border when ctrl+clicking
    public preventCtrlClickBorder($event) {
        if ($event.ctrlKey) {
            $event.preventDefault();
        }
    }

    public get addManyOptionsList() {
        return Array(20).fill(0).map((value, index) => index + 1);
    }

    // at this time @Input data is supposed to be already populated
    public ngOnInit() {
        this.model = this.pabTable.pab;
        this.refresh();
    }

    /** Unselects all selected text (side-effect of shift+clicking) */
    private clearSelection() {
        if (window.getSelection) {
            const sel = window.getSelection();
            sel.removeAllRanges();
        }
    }

    // extract PAB walls order
    private getSortedWallsUIDs(): string[] {
        const wallsUIDs: string[] = [];
        for (const c of this.pabTable.pab.children) {
            wallsUIDs.push(c.uid);
        }
        wallsUIDs.push(this.pabTable.pab.downWall.uid);
        return wallsUIDs;
    }

    /**
     * Ensures that this.selectedItems elements are ordered according to
     * the walls order in the PAB (important for interpolation)
     */
    private sortSelectedItems() {
        const wallsUIDs = this.getSortedWallsUIDs();
        // are items walls or devices ?
        if (this.onlyWallsAreSelected(false)) {
            // 1. walls : order by uid, according to model
            this.selectedItems.sort((a, b) => {
                const posA = wallsUIDs.indexOf(a.uid);
                const posB = wallsUIDs.indexOf(b.uid);
                return posA - posB;
            });
        } else {
            // 2. devices : order by parent (wall) uid, according to model
            this.selectedItems.sort((a, b) => {
                const posA = wallsUIDs.indexOf(a.parent.uid);
                const posB = wallsUIDs.indexOf(b.parent.uid);
                return posA - posB;
            });
        }
        return this.selectedItems;
    }

    /**
     * Builds the editable data grid from the Pab model
     */
    private refresh() {
        const maxNbDevices = this.findMaxNumberOfDevices();

        // adjuste precision once before anything else
        const nDigits = this.appSetupService.displayDigits;
        for (const c of this.model.children) {
            for (const p of c.parameterIterator) {
                if (p.visible) {
                    p.singleValue = round(p.singleValue, nDigits);
                }
            }
        }
        for (const p of this.model.downWall.parameterIterator) {
            if (p.visible) {
                p.singleValue = round(p.singleValue, nDigits);
            }
        }

        // 0. build spanned headers over real columns
        this.headers = [];
        // 1 header for basin
        let bs: any[] = this.model.children;
        bs = bs.concat(this.model.downWall);
        this.headers.push({
            title: this.i18nService.localizeText("INFO_PAB_BASSIN"),
            colspan: 6,
            selectable: bs
        });
        // 1 header for each device of the wall having the most devices (including downwall)
        for (let i = 0; i < maxNbDevices; i++) {
            this.headers.push({
                title: sprintf(this.i18nService.localizeText("INFO_PAB_CLOISON_OUVRAGE_N"), (i + 1)),
                colspan: 3,
                selectable: this.model.children.map(c => c.getChildren()[i]).concat(this.model.downWall.getChildren()[i]),
                selectableColumn: i
            });
        }

        // A. build columns set
        this.cols = [];
        // 6 cols for basin
        this.cols.push({
            title: this.i18nService.localizeText("INFO_PAB_NUM_BASSIN"),
            selectable: bs
        });
        this.cols.push({
            title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "LB"),
            selectable: bs
        });
        this.cols.push({
            title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "BB"),
            selectable: bs
        });
        this.cols.push({
            title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "QA"),
            selectable: bs
        });
        this.cols.push({
            title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "ZRMB"),
            selectable: bs
        });
        this.cols.push({
            title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "ZRAM"),
            selectable: bs
        });
        // no col for wall type (defined by rowspan-2 header above)
        // 3 cols for each device of the wall having the most devices (including downwall)
        for (let i = 0; i < maxNbDevices; i++) {
            this.cols.push({
                title: this.i18nService.localizeText("INFO_PAB_HEADER_TYPE"),
                selectable: this.model.children.map(c => c.getChildren()[i]).concat(this.model.downWall.getChildren()[i]),
                selectableColumn: i
            });
            this.cols.push({
                title: this.i18nService.localizeText("INFO_PAB_HEADER_PARAMETERS"),
                selectable: this.model.children.map(c => c.getChildren()[i]).concat(this.model.downWall.getChildren()[i]),
                selectableColumn: i
            });
            this.cols.push({
                title: this.i18nService.localizeText("INFO_PAB_HEADER_VALUES"),
                selectable: this.model.children.map(c => c.getChildren()[i]).concat(this.model.downWall.getChildren()[i]),
                selectableColumn: i
            });
        }

        // B. Build rows set
        this.rows = [];
        // B.1 many rows for each wall
        let childIndex = 0;
        for (const cloison of this.model.children) {
            // admissible LoiDebit
            const loisCloisons = cloison.getLoisAdmissiblesArray().map(l => { // @TODO move up ? (same for all cloisons)
                return {
                    label: this.i18nService.localizeLoiDebit(l),
                    value: l
                };
            });
            // as much rows as the greatest number of parameters among its devices
            const maxNbParams = this.findMaxNumberOfDeviceParameters(cloison);
            for (let i = 0; i < maxNbParams; i++) {
                // build device params row
                const deviceParamRow = { selectable: cloison, cells: [] };
                // basin number
                if (i === 0) {
                    deviceParamRow.cells.push({
                        value: childIndex + 1,
                        rowspan: maxNbParams + 1,
                        class: "basin_number",
                        selectable: cloison
                    });
                }
                // 5 empty cells
                if (i === 0) {
                    deviceParamRow.cells.push({
                        colspan: 5,
                        rowspan: maxNbParams,
                        selectable: cloison
                    });
                }
                // device param cells : 3 cells for each device
                for (const ouvrage of cloison.structures) {
                    const nvParam = ouvrage.getNthVisibleParam(i);
                    const nvParamTitle = nvParam ? this.formService.expandVariableNameAndUnit(CalculatorType.Pab, nvParam.symbol) : "";
                    // cell 1 : device type
                    if (i === 0) { // 1st row
                        deviceParamRow.cells.push({
                            model: ouvrage,
                            modelValue: ouvrage.properties.getPropValue("loiDebit"),
                            options: loisCloisons,
                            selectable: ouvrage
                        });
                    }
                    // fill space
                    if (i === 1) {
                        deviceParamRow.cells.push({
                            rowspan: (maxNbParams - 1),
                            selectable: ouvrage
                        });
                    }
                    // cell 2 : param name
                    if (nvParam) {
                        deviceParamRow.cells.push({
                            value: nvParam.symbol,
                            title: nvParamTitle,
                            selectable: ouvrage
                        });
                    } else {
                        deviceParamRow.cells.push({
                            selectable: ouvrage
                        });
                    }
                    // cell 3 : param value
                    if (nvParam) {
                        deviceParamRow.cells.push({
                            model: nvParam,
                            title: nvParamTitle,
                            selectable: ouvrage
                        });
                    } else {
                        deviceParamRow.cells.push({
                            selectable: ouvrage
                        });
                    }
                }
                // fill horizontal space
                const devDiff =  (maxNbDevices - cloison.structures.length);
                if (i === 0) {
                    for (let j = 0; j < devDiff; j++) {
                        deviceParamRow.cells.push({
                            colspan: 3,
                            rowspan: maxNbParams,
                            selectable: cloison,
                            selectableColumn: cloison.structures.length + j
                        });
                    }
                }
                // done !
                this.rows.push(deviceParamRow);
            }
            // 1 row for the basin after the wall
            const basinRow: { selectable: any, cells: any[] } = {
                selectable: cloison,
                cells: [
                    // no cell for basin number (defined by rowspan-n cell above)
                    {
                        model: cloison.prms.LB,
                        title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "LB")
                    },
                    {
                        model: cloison.prms.BB,
                        title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "BB")
                    },
                    {
                        model: cloison.prms.QA,
                        title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "QA")
                    },
                    {
                        model: cloison.prms.ZRMB,
                        title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "ZRMB")
                    },
                    {
                        model: cloison.prms.ZRAM,
                        title: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, "ZRAM")
                    }
                ]
            };
            // fill horizontal space
            for (let i = 0; i < maxNbDevices; i++) {
                basinRow.cells.push({
                    colspan: 3
                });
            }
            // done !
            this.rows.push(basinRow);
            childIndex ++;
        }

        // B.2 many rows for downwall
        // admissible LoiDebit
        const loisAval = this.model.downWall.getLoisAdmissiblesArray().map(l => {
            return {
                label: this.i18nService.localizeLoiDebit(l),
                value: l
            };
        });
        // as much rows as the greatest number of parameters among its devices
        const maxNbParamsDW = this.findMaxNumberOfDeviceParameters(this.model.downWall);
        for (let i = 0; i < maxNbParamsDW; i++) {
            // build device params row
            const deviceParamRowDW = { selectable: this.model.downWall, cells: [] };
            // basin number
            if (i === 0) {
                deviceParamRowDW.cells.push({
                    value: "Aval",
                    rowspan: maxNbParamsDW,
                    class: "basin_number",
                    selectable: this.model.downWall
                });
            }
            // 5 empty cells
            if (i === 0) {
                deviceParamRowDW.cells.push({
                    colspan: 5,
                    rowspan: maxNbParamsDW,
                    selectable: this.model.downWall
                });
            }
            // downwall device param cells : 3 cells for each device
            for (const ouvrage of this.model.downWall.structures) {
                const nvParam = ouvrage.getNthVisibleParam(i);
                const nvParamTitle = nvParam ? this.formService.expandVariableNameAndUnit(CalculatorType.Pab, nvParam.symbol) : "";
                // cell 1 : device type
                if (i === 0) { // 1st row
                    deviceParamRowDW.cells.push({
                        model: ouvrage,
                        modelValue: ouvrage.properties.getPropValue("loiDebit"),
                        options: loisAval
                    });
                }
                // fill space
                if (i === 1) {
                    deviceParamRowDW.cells.push({
                        rowspan: (maxNbParamsDW - 1),
                        selectable: ouvrage
                    });
                }
                // cell 2 : param name
                if (nvParam) {
                    deviceParamRowDW.cells.push({
                        value: nvParam.symbol,
                        title: nvParamTitle,
                        selectable: ouvrage
                    });
                } else {
                    deviceParamRowDW.cells.push({
                        selectable: ouvrage
                    });
                }
                // cell 3 : param value
                if (nvParam) {
                    deviceParamRowDW.cells.push({
                        model: nvParam,
                        title: nvParamTitle,
                        selectable: ouvrage
                    });
                } else {
                    deviceParamRowDW.cells.push({
                        selectable: ouvrage
                    });
                }
            }
            // fill horizontal space
            const devDiff =  (maxNbDevices - this.model.downWall.structures.length);
            if (i === 0) {
                for (let j = 0; j < devDiff; j++) {
                    deviceParamRowDW.cells.push({
                        colspan: 3,
                        rowspan: maxNbParamsDW,
                        selectable: this.model.downWall,
                        selectableColumn: this.model.downWall.structures.length + j
                    });
                }
            }
            // done !
            this.rows.push(deviceParamRowDW);
        }

        this.updateValidity();
    }

    private findMaxNumberOfDevices(): number {
        let maxNbDevices = 1;
        for (const w of this.model.children) {
            maxNbDevices = Math.max(maxNbDevices, w.getChildren().length);
        }
        maxNbDevices = Math.max(maxNbDevices, this.model.downWall.getChildren().length);
        return maxNbDevices;
    }

    private findMaxNumberOfDeviceParameters(struct: ParallelStructure): number {
        let maxNbParams = 1;
        for (const d of struct.getChildren()) {
            let nbParams = 0;
            for (const p of d.parameterIterator) {
                if (p.visible) {
                    // console.log("(counting)", p.symbol);
                    nbParams ++;
                }
            }
            // console.log(">>> child params: ", nbParams);
            maxNbParams = Math.max(maxNbParams, nbParams);
        }
        return maxNbParams;
    }

    /** returns true if exactly one device is selected, and nothing else */
    public get selectionIsOneDevice() {
        return (
            this.selectedItems.length === 1
            && this.selectedItem instanceof Structure
        );
    }

    /**
     * Returns true if there is at least one selected item,
     * and all selected items are devices
     */
    private onlyDevicesAreSelected() {
        let ok = false;
        if (this.selectedItems.length > 0) {
            ok = true;
            for (const s of this.selectedItems) {
                ok = ok && (s instanceof Structure);
            }
        }
        return ok;
    }

    /**
     * Returns true if there is at least one selected item,
     * all selected items are devices, and belong to the same column
     */
    private onlyDevicesOfTheSameColumnAreSelected() {
        let ok = false;
        let columnIndex: number;
        if (this.selectedItems.length > 0) {
            ok = true;
            for (const s of this.selectedItems) {
                if (s instanceof Structure) {
                    const ci = s.findPositionInParent();
                    ok = ok && (columnIndex === undefined || columnIndex === ci);
                    columnIndex = ci;
                } else {
                    ok = false;
                }
            }
        }
        return ok;
    }

    /**
     * Returns true if there is at least one selected item,
     * and all selected items are walls
     */
    private onlyWallsAreSelected(excludeDownwall: boolean = true) {
        let ok = false;
        if (this.selectedItems.length > 0) {
            ok = true;
            for (const s of this.selectedItems) {
                if (excludeDownwall) {
                    ok = ok && (s instanceof Cloisons);
                } else {
                    ok = ok && (s instanceof ParallelStructure);
                }
            }
        }
        return ok;
    }

    public get relatedEntityTitle() {
        let title = "";
        if (this.onlyDevicesAreSelected()) {
            title = "Ouvrages";
        } else if (this.onlyWallsAreSelected()) {
            title = "Cloisons";
        }
        if (title !== "") {
            title += " :";
        }
        return title;
    }

    public get enableAddButton() {
        return (
            this.onlyDevicesOfTheSameColumnAreSelected()
            || (
                this.selectedItems.length === 1
                && ! (this.selectedItem instanceof CloisonAval) // exclude downwall
            )
        );
    }

    public get enableCopyButton() {
        return this.enableAddButton;
    }

    public get enableUpButton() {
        return (
            this.selectedItems.length === 1
            && ! (this.selectedItem instanceof CloisonAval) // exclude downwall
            && this.selectedItem.parent
            && this.selectedItem.findPositionInParent() !== 0
        );
    }

    public get enableDownButton() {
        return (
            this.selectedItems.length === 1
            && ! (this.selectedItem instanceof CloisonAval) // exclude downwall
            && this.selectedItem.parent
            && this.selectedItem.findPositionInParent() < (this.selectedItem.parent.getChildren().length - 1)
        );
    }

    public get enableRemoveButton() {
        return (
            this.selectedItems.length === 1
            && ! (this.selectedItem instanceof CloisonAval) // exclude downwall
            && this.selectedItem.parent
            && this.selectedItem.parent.getChildren().length > 1
        );
    }

    /**
     * returns true if at least one object is selected
     */
    public get enableEditPabButton() {
        return (
            this.selectedItems.length > 0
            && (
                this.onlyDevicesAreSelected()
                || this.onlyWallsAreSelected(false)
            )
        );
    }

    public onAddClick() {
        // add default item
        for (let i = 0; i < this.childrenToAdd; i++) {
            for (const si of this.selectedItems) {
                if (si instanceof Structure) {
                    // add new default device for wall parent
                    const newDevice = Session.getInstance().createNub(
                        new Props({
                            calcType: CalculatorType.Structure,
                            loiDebit: (si.parent as ParallelStructure).getDefaultLoiDebit()
                        })
                    );
                    si.parent.addChild(newDevice, si.findPositionInParent());

                } else {
                    // add new default wall for PAB parent
                    const newWall = Session.getInstance().createNub(
                        new Props({
                            calcType: CalculatorType.Cloisons
                        })
                    );
                    // add new default device for new wall
                    const newDevice = Session.getInstance().createNub(
                        new Props({
                            calcType: CalculatorType.Structure,
                            loiDebit: (newWall as ParallelStructure).getDefaultLoiDebit()
                        })
                    );
                    newWall.addChild(newDevice);
                    this.model.addChild(newWall, si.findPositionInParent());
                }
            }
        }
        this.refresh();

        // notify
        const pos = this.selectedItem.findPositionInParent() + 1;
        let msg: string;
        if (this.childrenToAdd === 1 && this.selectedItems.length === 1) {
            if (this.selectedItem instanceof Structure) {
                msg = this.i18nService.localizeText("INFO_DEVICE_ADDED");
            } else {
                msg = this.i18nService.localizeText("INFO_WALL_ADDED");
            }
        } else {
            const size = (this.childrenToAdd * this.selectedItems.length);
            if (this.selectedItem instanceof Structure) {
                msg = sprintf(this.i18nService.localizeText("INFO_DEVICE_ADDED_N_TIMES"), size);
            } else {
                msg = sprintf(this.i18nService.localizeText("INFO_WALL_ADDED_N_TIMES"), size);
            }
        }
        this.notifService.notify(msg);

        this.childrenToAdd = 1; // reinit to avoid confusion
    }

    public onCopyClick() {
        // cloned selected item
        for (let i = 0; i < this.childrenToAdd; i++) {
            for (const si of this.selectedItems) {
                const newChild = Session.getInstance().createNub(
                    si.properties.clone(),
                    si.parent
                );
                // copy parameter values
                for (const p of si.prms) {
                    newChild.getParameter(p.symbol).singleValue = p.singleValue;
                }
                // copy children
                if (si instanceof ParallelStructure) {
                    for (const c of si.getChildren()) {
                        const newGrandChild = Session.getInstance().createNub(
                            c.properties.clone(),
                            newChild
                        );
                        // copy children parameters values
                        for (const p of c.prms) {
                            newGrandChild.getParameter(p.symbol).singleValue = p.singleValue;
                        }
                        // add to parent
                        newChild.addChild(
                            newGrandChild,
                            c.findPositionInParent()
                        );
                    }
                }
                // add to parent
                si.parent.addChild(
                    newChild,
                    si.findPositionInParent()
                );
            }
        }
        this.refresh();

        // notify
        const pos = this.selectedItem.findPositionInParent() + 1;
        let msg: string;
        if (this.childrenToAdd === 1 && this.selectedItems.length === 1) {
            if (this.selectedItem instanceof Structure) {
                msg = sprintf(this.i18nService.localizeText("INFO_DEVICE_COPIED"), pos);
            } else {
                msg = sprintf(this.i18nService.localizeText("INFO_WALL_COPIED"), pos);
            }
        } else {
            const size = (this.childrenToAdd * this.selectedItems.length);
            if (this.selectedItem instanceof Structure) {
                msg = sprintf(this.i18nService.localizeText("INFO_DEVICE_COPIED_N_TIMES"), pos, size);
            } else {
                msg = sprintf(this.i18nService.localizeText("INFO_WALL_COPIED_N_TIMES"), pos, size);
            }
        }
        this.notifService.notify(msg);

        this.childrenToAdd = 1; // reinit to avoid confusion
    }

    public onMoveUpClick() {
        const pos = this.selectedItem.findPositionInParent() + 1;
        this.selectedItem.parent.moveChildUp(this.selectedItem);
        if (this.selectedItem instanceof Structure) {
            this.notifService.notify(sprintf(this.i18nService.localizeText("INFO_DEVICE_MOVED"), pos));
        } else {
            this.notifService.notify(sprintf(this.i18nService.localizeText("INFO_WALL_MOVED"), pos));
        }
        this.refresh();
    }

    public onMoveDownClick() {
        const pos = this.selectedItem.findPositionInParent() + 1;
        this.selectedItem.parent.moveChildDown(this.selectedItem);
        if (this.selectedItem instanceof Structure) {
            this.notifService.notify(sprintf(this.i18nService.localizeText("INFO_DEVICE_MOVED"), pos));
        } else {
            this.notifService.notify(sprintf(this.i18nService.localizeText("INFO_WALL_MOVED"), pos));
        }
        this.refresh();
    }

    public onRemoveClick() {
        const pos = this.selectedItem.findPositionInParent() + 1;
        this.selectedItem.parent.deleteChild(this.selectedItem.findPositionInParent());
        if (this.selectedItem instanceof Structure) {
            this.notifService.notify(sprintf(this.i18nService.localizeText("INFO_DEVICE_REMOVED"), pos));
        } else {
            this.notifService.notify(sprintf(this.i18nService.localizeText("INFO_WALL_REMOVED"), pos));
        }
        this.selectedItems = [];
        this.refresh();
    }

    public get uitextAdd(): string {
        return this.i18nService.localizeText("INFO_FIELDSET_ADD");
    }

    public get uitextCopy(): string {
        return this.i18nService.localizeText("INFO_FIELDSET_COPY");
    }

    public get uitextRemove(): string {
        return this.i18nService.localizeText("INFO_FIELDSET_REMOVE");
    }

    public get uitextMoveUp(): string {
        if (this.selectionIsOneDevice) {
            return this.i18nService.localizeText("INFO_FIELDSET_MOVE_LEFT");
        } else {
            return this.i18nService.localizeText("INFO_FIELDSET_MOVE_UP");
        }
    }

    public get uitextMoveDown(): string {
        if (this.selectionIsOneDevice) {
            return this.i18nService.localizeText("INFO_FIELDSET_MOVE_RIGHT");
        } else {
            return this.i18nService.localizeText("INFO_FIELDSET_MOVE_DOWN");
        }
    }

    /** Replace device Nub when LoiDebit is changed */
    public loiDebitSelected($event: any, cell: any) {
        const device = cell.model as Nub;
        // create new child device
        const newDevice = Session.getInstance().createNub(
            new Props({
                calcType: CalculatorType.Structure,
                loiDebit: $event.value
            })
        );
        // replace the current one
        device.parent.replaceChildInplace(device, newDevice);
        this.refresh();
    }

    // show modal dialog for values edition
    public showEditPab() {
        if (this.selectedItems.length > 0) {

            // list variables eligible to modification
            const availableVariables: { label: string, value: string, occurrences: number, first: number, last: number }[] = [];
            for (const c of this.selectedItems) {
                for (const p of c.parameterIterator) { // deep one
                    if (
                        p.visible &&
                        ! availableVariables.map(av => av.value).includes(p.symbol)
                    ) {
                        availableVariables.push({
                            label: this.formService.expandVariableNameAndUnit(CalculatorType.Pab, p.symbol),
                            value: p.symbol,
                            occurrences: 0,
                            first: undefined,
                            last: undefined
                        });
                    }
                }
            }
            // find their min/max values (2nd pass)
            for (const av of availableVariables) {
                for (const c of this.selectedItems) {
                    for (const p of c.parameterIterator) {
                        if (p.visible && p.symbol === av.value) {
                            av.occurrences ++;
                            if (av.first === undefined) {
                                av.first = p.singleValue;
                            }
                            av.last = p.singleValue;
                        }
                    }
                }
            }

            // sum up selected items
            const walls: ParallelStructure[] = [];
            const wallsDevices: Structure[] = [];
            const devices: Structure[] = [];
            let vertical = true; // @TODO vertical AND consecutive !
            let firstDevicePosition: number;
            // 1st pass
            for (const s of this.selectedItems) {
                if (s instanceof ParallelStructure) {
                    walls.push(s);
                    for (const c of s.structures) {
                        if (firstDevicePosition === undefined) {
                            firstDevicePosition = c.findPositionInParent();
                        } else {
                            vertical = (vertical && (c.findPositionInParent() === firstDevicePosition));
                        }
                        wallsDevices.push(c);
                    }
                }
            }
            // 2nd pass
            for (const c of this.selectedItems) {
                if (c instanceof Structure) {
                    if (! wallsDevices.includes(c)) {
                        if (firstDevicePosition === undefined) {
                            firstDevicePosition = c.findPositionInParent();
                        } else {
                            vertical = (vertical && (c.findPositionInParent() === firstDevicePosition));
                        }
                        devices.push(c);
                    }
                }
            }

            // open dialog
            const dialogRef = this.editPabDialog.open(
                DialogEditPabComponent,
                {
                    data: {
                        availableVariables: availableVariables,
                        selectedItemsAbstract: {
                            walls: walls.length,
                            wallsDevices: wallsDevices.length,
                            devices: devices.length
                        },
                        vertical: vertical // used to enable interpolation
                    },
                    disableClose: true
                }
            );

            // apply modifications
            dialogRef.afterClosed().subscribe(result => {
                if (result) {
                    /* console.log("Apply values in parent !!", result.action, result.variable, result.value,
                      result.delta, result.variableDetails); */
                    switch (result.action) {
                        case "set-value":
                            for (const s of this.selectedItems) {
                                for (const p of s.parameterIterator) { // deep
                                    if (p.symbol === result.variable) {
                                        p.singleValue = result.value;
                                    }
                                }
                            }
                            break;

                        case "delta":
                            for (const s of this.selectedItems) {
                                for (const p of s.parameterIterator) { // deep
                                    if (p.symbol === result.variable) {
                                        p.singleValue += result.delta;
                                    }
                                }
                            }
                            break;

                        case "interpolate":
                            if (result.variableDetails.occurrences > 1) {
                                const nDigits = this.appSetupService.displayDigits;
                                // build values list
                                const interpolatedValues: number[] = [];
                                const step = (
                                    (result.variableDetails.last - result.variableDetails.first)
                                    / (result.variableDetails.occurrences - 1)
                                );
                                interpolatedValues.push(result.variableDetails.first);
                                let currentValue: number = result.variableDetails.first;
                                for (let i = 0; i < result.variableDetails.occurrences - 2; i++) {
                                    currentValue += step;
                                    interpolatedValues.push(round(currentValue, nDigits));
                                }
                                interpolatedValues.push(result.variableDetails.last);
                                // apply
                                let idx = 0;
                                for (const s of this.selectedItems) {
                                    for (const p of s.parameterIterator) { // deep
                                        if (p.symbol === result.variable) {
                                            p.singleValue = interpolatedValues[idx];
                                            idx ++;
                                        }
                                    }
                                }
                            } else {
                                throw new Error(
                                    `showEditPab() : cannot interpolate, too few occurrences (${result.variableDetails.occurrences})`
                                );
                            }
                            break;
                    }
                }
            });
        }
    }

    public ngAfterViewInit() {
        this.updateValidity();
    }

    /**
     * Computes the global Pab validity : validity of every cell of every row
     */
    private updateValidity() {
        this._isValid = true;
        for (const r of this.rows) {
            for (const c of r.cells) {
                this._isValid = this._isValid && ! this.isInvalid(c);
            }
        }
        this.validChange.emit();
    }

    public get uitextEditPabTable() {
        return this.i18nService.localizeText("INFO_PAB_EDIT_VALUES");
    }
}
